#include "SolutionPopulationC.h"
#include "../Common/ArrayC.t"
#include "EncounterC.h"
#include "PopulationPoolC.t"


extern MLCG RandomGenerator;


SolutionPopulationC::SolutionPopulationC
(
	UnsignedIntegerT const iSizeOfSolutionPopulation,
	UnsignedIntegerT const iSizeOfOffspring,
	UnsignedIntegerT const iSizeOfGene,
	UnsignedIntegerT const iSizeOfFitnessHistory,
	ZeroToOneRealRealT const iNormalMutationProbability,
	ZeroToOneRealRealT const iEqualMutationProbability,
	SelectionC * iSelection,
	RandomCspC * iProblem
)
{
	UnsignedIntegerT i;
	SolutionGenotypeC * SolutionGenotype;

	SizeOfSolutionPopulation = iSizeOfSolutionPopulation;
	SizeOfOffspring = iSizeOfOffspring;
	SizeOfGene = iSizeOfGene;
	SizeOfFitnessHistory = iSizeOfFitnessHistory;

	EqualMutationProbability = iEqualMutationProbability;
	NormalMutationProbability = iNormalMutationProbability;

	Problem = iProblem;
	Selection = iSelection;

	Pool = new PopulationPoolC <SolutionGenotypeC *> (SizeOfSolutionPopulation + SizeOfOffspring);

	Parents = new ArrayC <SolutionGenotypeC *> (SizeOfOffspring);
	Offspring = new ArrayC <SolutionGenotypeC *> (SizeOfOffspring);

	GenerationNumber = 0;
	NumberOfGeneratedGenotypes = 0;
	FoundSolution = false;

	for (i = 0; i < SizeOfSolutionPopulation; i++)
	{
		SolutionGenotype = new SolutionGenotypeC (SizeOfGene, SizeOfFitnessHistory, Problem);
		Pool->Insert (SolutionGenotype);
	}

} // SolutionPopulationC::SolutionPopulationC



SolutionPopulationC::~SolutionPopulationC ()
{
	Pool->First ();
	while (!Pool->IsEnd ())
	{
		Pool->SetBookmark ();
		delete Pool->Current ();
		Pool->Next ();
		Pool->DeleteBookmark ();
	}
	
	delete Pool;
	delete Parents;
	delete Offspring;
} // SolutionPopulationC::~SolutionPopulationC



void const SolutionPopulationC::Print ()
{
	Pool->First ();
	while (!Pool->IsEnd ())
	{
		Pool->Current ()->Print ();
		cout
			<< setw(4) << "<" << Pool->Current ()->GetFitnessValue () << ">"
			<< setw(5) << "[" << Pool->Current ()->GetNumberOfUnsatisfiedConstraints () << "]"
			<< "\n"
		;
		Pool->Next ();
	}
	
} // SolutionPopulationC::Print



void const SolutionPopulationC::PrintBestGenotype ()
{
	Pool->First ();
	Pool->Current ()->Print ();
	
} // SolutionPopulationC::PrintBestGenotype



void const SolutionPopulationC::InitialiseFitness
(
	EncounterC * iEncounter
)
{
	Encounter = iEncounter;
	UnsignedIntegerT i;
	ArrayC <SolutionGenotypeC *> TempPool (SizeOfSolutionPopulation);
	
	i = 0;
	Pool->First ();
	while (!Pool->IsEnd ())
	{
		Pool->SetBookmark ();
		TempPool [i] = Pool->Current ();
		TempPool [i]->InitialiseFitness (Encounter);
		i++;
		Pool->Next ();
		Pool->DeleteBookmark ();
	}

	for (i = 0; i < SizeOfSolutionPopulation; i++)
	{
		Pool->Insert (TempPool [i]);
	}

} // SolutionPopulationC::InitialiseFitness



void const SolutionPopulationC::DoOneGeneration ()
{
	SelectParents ();

	Reproduce ();

	AddOffspring ();

	GenerationNumber++;

} // SolutionPopulationC::DoOneGeneration



void const SolutionPopulationC::SelectParents ()
{
	UnsignedIntegerT i;

	Pool->First ();
	for (i = 0; i < SizeOfOffspring; i++)
	{
		(*Parents) [i] = Pool->Current ();
		Pool->Next ();
	}

} // SolutionPopulationC::SelectParents



/*
 * Does a complete production step of the population; producing offspring with
 * crossover, mutating the offspring, checking for a solution initialising
 * fitness of offspring.
 *
 * Comment: could be faster if we assume that the crossover always returns two
 * different genotypes, unless the two parents are equal. To accomplish this
 * the crossover should return true if it succeeded in making new offspring or
 * if it just made two copies. The chance of two different parents to produce
 * the same children should be very small using 'two point surrogate parents
 * crossover'.
 *
 */

void const SolutionPopulationC::Reproduce ()
{
	UnsignedIntegerT i;

	NumberOfGeneratedGenotypes += SizeOfOffspring;
	for (i = 0; i < SizeOfOffspring; i++)
	{
		(*Offspring) [i] = new SolutionGenotypeC (SizeOfGene, SizeOfFitnessHistory, Problem);
	}

	for (i = 0; i < SizeOfOffspring; i += 2)
	{
		(*Parents) [i]->Crossover (*((*Parents) [i+1]), *((*Offspring) [i]), *((*Offspring) [i+1]));
		if (*(*Offspring) [i] == *(*Offspring) [i +1])
		{
			(*Offspring) [i]->Mutate (EqualMutationProbability);
			(*Offspring) [i]->Mutate (EqualMutationProbability);
		}
		else
		{
			(*Offspring) [i]->Mutate (NormalMutationProbability);
			(*Offspring) [i]->Mutate (NormalMutationProbability);
		}
	}

	for (i = 0; i < SizeOfOffspring; i++)
	{
		if ((*Offspring) [i]->GetNumberOfUnsatisfiedConstraints () == 0)
		{
			FoundSolution = true;
			FoundSolutionGenotype = (*Offspring) [i];
		}
	}

	for (i = 0; i < SizeOfOffspring; i++)
	{
		(*Offspring) [i]->InitialiseFitness (Encounter);
	}

} // SolutionPopulationC::Reproduce



void const SolutionPopulationC::AddOffspring ()
{
	UnsignedIntegerT i;

	for (i = 0; i < SizeOfOffspring; i++)
	{
		Pool->Insert ((*Offspring) [i]);
	}

	Pool->Last ();
	for (i = 0; i < SizeOfOffspring; i++)
	{
		Pool->SetBookmark ();
		delete Pool->Current ();
		Pool->Previous ();
		Pool->DeleteBookmark ();
	}

} // SolutionGenotypeC::AddOffspring



SolutionFitnessT const SolutionPopulationC::GetFitnessOfBestGenotype ()
{
	Pool->First ();

	return (Pool->Current ()->GetFitnessValue ());

} // SolutionPopulationC::GetFitnessOfBestSolutionGenotype



SolutionFitnessT const SolutionPopulationC::GetNumberOfUnsatisfiedConstraintsOfBestGenotype ()
{
	Pool->First ();

	return (Pool->Current ()->GetNumberOfUnsatisfiedConstraints ());

} // SolutionPopulationC::GetNumberOfUnsatisfiedConstraintsOfBestGenotype



UnsignedIntegerT const SolutionPopulationC::GetGenerationNumber () const
{
	return (GenerationNumber);

} // SolutionPopulationC::GenerationNumber



void const SolutionPopulationC::InjectGenotype
(
	SolutionGenotypeC * iSolutionGenotype
)
{
	Pool->Insert (iSolutionGenotype);

} // SolutionGenotypeC::InjectGenotype



SolutionGenotypeC * SolutionPopulationC::FetchGenotype ()
{
	UnsignedIntegerT i;
	SolutionGenotypeC * SolutionGenotype;

	Pool->First ();
	for (i = 0; i < Selection->GetIndex (); i++)
	{
		Pool->Next ();
	}
	SolutionGenotype = Pool->Current ();
	Pool->SetBookmark ();
	Pool->None ();
	Pool->DeleteBookmark ();

	return (SolutionGenotype);

} // SolutionPopulationC::FetchGenotype


SolutionGenotypeC * SolutionPopulationC::GetFoundSolutionGenotype ()
{
	if (!GetFoundSolution ())
	{
		cerr << "Warning: asking for a solution that has not been found yet.\n";
		return (0);
	}
	else
		return (FoundSolutionGenotype);
}

BoolT const SolutionPopulationC::GetFoundSolution () const
{
	return (FoundSolution);

} // SolutionPopulationC::GetFoundSolution



UnsignedIntegerT const SolutionPopulationC::GetNumberOfEncountersDone () const
{
	return (Encounter->GetTotalNumberOfEncounters ());

} // SolutionPopulationC::GetNumberOfFitnessEvaluation



UnsignedIntegerT const SolutionPopulationC::GetNumberOfGeneratedGenotypes () const
{
	return (NumberOfGeneratedGenotypes);

} // SolutionPopulationC::GetNumberOfGeneratedGenotypes



UnsignedIntegerT const SolutionPopulationC::GetNumberOfGenerations () const
{
	return (GenerationNumber);

} // SolutionPopulationC::GetNumberOfGenerations


// eof SolutionPopulationC.cc
